@@ -128,7 +128,8 @@ static void *nativeList;
 static void *nativePtrList;
 static void *toptrList;
 static void *parmList;
-#define NUM_CONFIG_CMDS 6
+static void *charList;
+#define NUM_CONFIG_CMDS 7
 static SYConfigCmds configcmds[NUM_CONFIG_CMDS+1];
 
 /* We also need to make some edits to the types occasionally.  First, note
@@ -184,6 +185,7 @@ int GetTypeName ( FILE *, FILE *, TYPE_LIST *, int, int, int );
 int GetArgName ( FILE *, FILE *, ARG_LIST *, TYPE_LIST *, int );
 void FixupArgNames( int, ARG_LIST * );
 void OutputBalancedString ( FILE *, FILE *, int, int );
+void SkipBalancedString ( FILE *, FILE *, int, int );
 char *ToCPointer ( char *, char *, int );
 const char *ArgToFortran( const char *typeName );
 void FreeArgs( ARG_LIST *, int );
@@ -193,8 +195,8 @@ void Abort( const char *, const char *, int );
 void DoBfortHelp ( char * );
 
 int stringDecl(FILE *fout, int nargs, ARG_LIST args[]);
-int stringFtoC(FILE *fout, int nargs, ARG_LIST args[]);
-int stringCtoF(FILE *fout, int nargs, ARG_LIST args[]);
+int stringFtoC(FILE *fout, int nargs, ARG_LIST args[], TYPE_LIST types[]);
+int stringCtoF(FILE *fout, int nargs, ARG_LIST args[], TYPE_LIST types[]);
 
 #define ABORT(msg) Abort(msg,__FILE__,__LINE__)
 
@@ -303,6 +305,7 @@ int main( int argc, char **argv )
     SYConfigDBInit("nativeptr", &nativePtrList);
     SYConfigDBInit("toptr", &toptrList);
     SYConfigDBInit("parm", &parmList);
+    SYConfigDBInit("char", &charList);
     configcmds[0].name     = "native";
     configcmds[0].docmd    = SYConfigDBInsert;
     configcmds[0].cmdextra = nativeList;
@@ -315,10 +318,13 @@ int main( int argc, char **argv )
     configcmds[3].name     = "parm";
     configcmds[3].docmd    = SYConfigDBInsert;
     configcmds[3].cmdextra = parmList;
+    configcmds[4].name     = "char";
+    configcmds[4].docmd    = SYConfigDBInsert;
+    configcmds[4].cmdextra = charList;
     /* These next two are used by bfort2 but ignore by bfort */
-    configcmds[4].name     = "fromptr";
-    configcmds[4].docmd    = SYConfigDBIgnore;
-    configcmds[4].cmdextra = 0;
+    //    configcmds[4].name     = "fromptr";
+    //configcmds[4].docmd    = SYConfigDBIgnore;
+    //configcmds[4].cmdextra = 0;
     configcmds[5].name     = "declptr";
     configcmds[5].docmd    = SYConfigDBIgnore;
     configcmds[5].cmdextra = 0;
@@ -399,9 +405,7 @@ int main( int argc, char **argv )
 	exit( 1 );
     }
     if (SYArgHasName( &argc, argv, 1, "-version" )) {
-	printf( "bfort (sowing) release %d.%d.%d of %s\n",
-		PATCHLEVEL, PATCHLEVEL_MINOR, PATCHLEVEL_SUBMINOR,
-		PATCHLEVEL_RELEASE_DATE );
+	printf( "bfort (sowing) release %s of %s\n", PACKAGE_VERSION, PATCHLEVEL_RELEASE_DATE);
 	exit( 1 );
     }
 
@@ -451,9 +455,10 @@ int main( int argc, char **argv )
 	if (!errArgNameLocal) {
 	    if (SYConfigDBLookup("parm", "errparmlocal",
 				 &errArgNameLocal, parmList) != 1) {
-		errArgNameLocal = "__ierr";
+		errArgNameLocal = "ierr";
 	    }
 	}
+        errArgNameLocal = "ierr";
     }
 
     /* Get the name to use for the length of the string argument */
@@ -742,6 +747,7 @@ void OutputRoutine( FILE *fin, FILE *fout, char *name, char *filename,
     int         ntypes;
     int         flag2 = 0, cerr;
 
+    printf("output routine name %s\n",name);
     /* Check to see if this is a C-only routine */
     if (GetSubClass(&cerr) == 'C' && cerr == 0) {
 	if (!NoFortMsgs && !NoFortWarnings) {
@@ -1029,6 +1035,10 @@ void ProcessFunctionType( FILE *fin, FILE *fout, char *filename,
     in_args     = 0;
     p           = rcall;
     c           = FindNextANToken( fin, p, &nsp );
+
+    if (strcmp(p,"static") == 0) c           = FindNextANToken( fin, p, &nsp );
+    if (strcmp(p,"inline") == 0) c           = FindNextANToken( fin, p, &nsp );
+
 /*
    We check for routines that return (functions) versus ones that don't
    by looking for "void".  A special case is functions that return
@@ -1213,6 +1223,7 @@ void ProcessArgList( FILE *fin, FILE *fout, char *filename, int *is_function,
 	    break;
 	}
 	args[nargs].type = ntypes-1;
+
 	if (curtype) {
 	    /* Propagate information about the type to the specific argument */
 	    if (curtype->type_has_star)
@@ -1228,6 +1239,7 @@ void ProcessArgList( FILE *fin, FILE *fout, char *filename, int *is_function,
 		nstrings++;
 	    }
 	}
+
 	if (nargs >= MAX_ARGS) {
 	    ErrCnt++;
 	    fprintf( stderr, "Too many arguments to function %s; only %d supported and function has %d\n",
@@ -1254,7 +1266,7 @@ void ProcessArgList( FILE *fin, FILE *fout, char *filename, int *is_function,
 			     stringStyle == STRING_UNKNOWN)) {
 		int i;
 		for (i=0; i<nstrings; i++) {
-		    fprintf(fout, ", int %s%d", stringArgLenName, i);
+		    fprintf(fout, ", PETSC_FORTRAN_CHARLEN_T %s%d", stringArgLenName, i);
 		}
 	    }
 	    fputs( ")\n", fout );
@@ -1485,6 +1497,7 @@ void PrintBody( FILE *fout, int is_function, char *name, int nstrings,
 		int nargs, ARG_LIST *args, TYPE_LIST *types, RETURN_TYPE *rt )
 {
     int  i, j, nstr;
+    int  founddestroy = 0;
 
     fputs( "{\n", fout );
     /* Look for special-cases (FILE and declarations for string processing) */
@@ -1495,8 +1508,26 @@ void PrintBody( FILE *fout, int is_function, char *name, int nstrings,
     }
     stringDecl(fout, nargs, args);
 
+    if (nargs == 1 && args[0].has_star == 2) {
+        int len = (int) strlen(name);
+        if (len > 7) {
+            if (strcmp(name + len - 7, "Destroy") == 0) {
+           	fprintf(fout, "PETSC_FORTRAN_OBJECT_F_DESTROYED_TO_C_NULL(%s);\n ",args[0].name);
+                founddestroy = 1;
+            }
+        }
+    }
+    if (nargs == 2 && args[1].has_star == 2) {
+        int len = (int) strlen(name);
+        if (len > 6) {
+            if (strcmp(name + len - 6, "Create") == 0) {
+             	fprintf(fout, "PETSC_FORTRAN_OBJECT_CREATE(%s);\n ",args[1].name);
+            }
+        }
+    }
+
     /* Add code for the string processing */
-    stringFtoC(fout, nargs, args);
+    stringFtoC(fout, nargs, args, types);
 
     /* Generate the routine call with the return */
     if (is_function) {
@@ -1522,10 +1553,16 @@ void PrintBody( FILE *fout, int is_function, char *name, int nstrings,
 	if (args[i].is_FILE)
 	    fprintf( fout, "_fp%d", i );
 	else if (args[i].is_char && stringStyle == STRING_LENEND) {
-	    fprintf(fout, "_%stmp%d", stringArgLenName, nstr);
+            if (args[i].has_star == 1)
+	      fprintf(fout, "_%stmp%d", stringArgLenName, nstr);
+            else if (args[i].has_star == 0)
+	      fprintf(fout, "*%s", args[i].name);
+            else
+	      fprintf(fout, "(const char **)&_%stmp%d", stringArgLenName, nstr);
 	    nstr++;
-	}
-	else if (!args[i].is_native && args[i].has_star
+	} else if (strcmp(types[args[i].type].type,"PetscViewer") == 0 && args[i].has_star == 1) {
+         fprintf( fout, "PetscPatchDefaultViewers((PetscViewer*)%s)", args[i].name);
+	} else if (!args[i].is_native && args[i].has_star
 		 && !args[i].void_function) {
 	    if (args[i].has_star == 1 || !MultipleIndirectAreInts)
 		fprintf( fout, "%s",
@@ -1556,7 +1593,10 @@ void PrintBody( FILE *fout, int is_function, char *name, int nstrings,
     }
     fputs( ");\n", fout);
     /* May need to process strings on exit if argument was not const */
-    stringCtoF(fout, nargs, args);
+    stringCtoF(fout, nargs, args, types);
+
+    if (founddestroy) fprintf(fout, "if (*ierr) return;\nPETSC_FORTRAN_OBJECT_C_NULL_TO_F_DESTROYED(%s);\n ",args[0].name);
+
     fputs("}\n", fout );
 }
 
@@ -1691,7 +1731,7 @@ void PrintDefinition( FILE *fout, int is_function, char *name, int nstrings,
 		      int nargs, ARG_LIST *args, TYPE_LIST *types,
 		      RETURN_TYPE *rt )
 {
-    int  i;
+    int  i, petscobject_args = 0;
     const char *token = 0;
     char sname[2];      /* Use for short argnames, if enabled */
 
@@ -1723,6 +1763,18 @@ void PrintDefinition( FILE *fout, int is_function, char *name, int nstrings,
 	return;
     }
 
+    for (i=0; i<nargs; i++) {
+	if (strcmp(types[args[i].type].type,"PetscObject") == 0) petscobject_args++;
+	if (strcmp(types[args[i].type].type,"MatInfo") == 0 || strcmp(types[args[i].type].type,"MatFactorInfo") == 0 ||
+            strcmp(types[args[i].type].type,"DMDALocalInfo") == 0 ) {
+            args[i].has_array = 1;
+        }
+    }
+
+    if (petscobject_args) {
+      fprintf( fout, "#if defined(PETSC_HAVE_FORTRAN_TYPE_STAR)\n" );
+    }
+
     curCol = 0;
     /* Output the function definition */
     if (useFerr) {
@@ -1787,8 +1839,14 @@ void PrintDefinition( FILE *fout, int is_function, char *name, int nstrings,
 	    OutputFortranToken( fout, 7, "external" );
 	}
 	else {
+            if (args[i].is_char) {
+	    OutputFortranToken( fout, 7, ArgToFortran( "char*" ) );
+            } else if (strcmp(types[args[i].type].type,"PetscObject") == 0) {
+                OutputFortranToken( fout, 7, "type(*)");
+            } else {
 	    OutputFortranToken( fout, 7,
 				ArgToFortran( types[args[i].type].type ) );
+            }
 	}
 	if (useShortNames) {
 	    sname[0] = 'a' + (char)(i);
@@ -1874,6 +1932,10 @@ void PrintDefinition( FILE *fout, int is_function, char *name, int nstrings,
     OutputFortranToken( fout, 1, token );
     OutputFortranToken( fout, 1, name );
     OutputFortranToken( fout, 0, "\n" );
+    if (petscobject_args) {
+      fprintf( fout, "#endif\n" );
+    }
+
 }
 
 int NameHasUnderscore( char *p )
@@ -2156,10 +2218,17 @@ int GetTypeName( FILE *fin, FILE *fout, TYPE_LIST *type, int is_macro,
 	    ABORT( "Cannot append token to typename" );
 	}
 	/* Look for known names */
-	if (strcmp( token, "char" ) == 0) {
+	if (strcmp( token, "char" ) == 0){
 	    DBG("Saw char\n");
 	    type->is_char = 1;
 	}
+      	if (SYConfigDBLookup("char", token, NULL, charList) == 1) {
+            //if (strcmp( token, "PCType" ) == 0 || strcmp( token, "KSPType" ) == 0) {
+	    DBG("Saw char\n");
+	    type->is_char = 1;
+            type->type_has_star = 1;
+            strcpy(token,"char");
+	}
 	if (strcmp( token, "FILE" ) == 0) {
 	    DBG("Saw FILE\n");
 	    type->is_FILE = 1;
@@ -2436,12 +2505,12 @@ int GetArgName( FILE *fin, FILE *fout, ARG_LIST *arg, TYPE_LIST *type,
 	    nparen--;
 	else
 	    break;
-	OutputToken( fout, p, nsp );
+	if (!type->is_char || c != '*' || nstar < 2) OutputToken( fout, p, nsp );
 	c = FindNextANToken( fin, p, &nsp );
     }
 
     /* Current token is name */
-    arg->has_star = (nstar > 0);
+    if (type->is_char) arg->has_star = nstar; else arg->has_star = (nstar > 0);
     arg->name     = (char *)MALLOC( strlen(p) + 1 );
     if (MPIU_Strncpy( arg->name, p, strlen(p) + 1) ) {
 	ABORT( "Cannot set argument name" );
@@ -2477,6 +2546,7 @@ int GetArgName( FILE *fin, FILE *fout, ARG_LIST *arg, TYPE_LIST *type,
 	OutputToken( fout, "*", nsp );
 	nsp = 0;
     }
+
     if (nstar > 0) {
 	if (!NoFortMsgs && /* !is_function &&  */
 	    nstar > type->implied_star + 1) {
@@ -2493,11 +2563,12 @@ int GetArgName( FILE *fin, FILE *fout, ARG_LIST *arg, TYPE_LIST *type,
     /* Now, get the value from the peak above */
     nsp = nsp1;
     while (has_extra_chars && (c == '(' || (nparen > 0 && c == ')') || c == '[')) {
-	OutputToken( fout, p, nsp );
+	if (c != '[') OutputToken( fout, p, nsp );
 	if (c == '(') { OutputBalancedString( fin, fout, '(', ')' );
 	arg->void_function = 1; }
 	else if (c == ')') nparen--;
-	else if (c == '[') { OutputBalancedString( fin, fout, '[', ']' );
+	else if (c == '[') {if (arg->has_star && type->is_char) SkipBalancedString( fin, fout, '[', ']' ); else {
+                OutputToken( fout, p, nsp ); OutputBalancedString( fin, fout, '[', ']' );}
 	arg->has_array = 1; arg->has_star++; }
 	else if (c == ']') nbrack--;
 	else if (c == '*') ;
@@ -2529,6 +2600,18 @@ void OutputBalancedString( FILE *fin, FILE *fout, int cs, int ce )
     }
 }
 
+void SkipBalancedString( FILE *fin, FILE *fout, int cs, int ce )
+{
+    int c, bcount;
+    bcount = 1;
+    while (bcount) {
+	c = getc( fin );
+	if (c == cs) bcount++;
+	else if (c == ce) bcount--;
+    }
+}
+
+
 void DoBfortHelp( char *pgm )
 {
 fprintf( stderr, "%s - write a Fortran interface to C routines with\n",
@@ -2601,6 +2684,7 @@ const char *ArgToFortran( const char *typeName )
     const char *outName = 0;
     if (strcmp( typeName, "int") == 0) outName = "integer";
     else if (strcmp( typeName, "char" ) == 0) outName   = "character";
+    else if (strcmp( typeName, "char*" ) == 0) outName   = "character(*)";
     else if (strcmp( typeName, "double" ) == 0) outName = "double precision";
     else if (strcmp( typeName, "float" ) == 0) outName  = "real";
     else if (strcmp( typeName, "short" ) == 0) outName  = "integer*2";
@@ -2786,31 +2870,46 @@ int stringDecl(FILE *fout, int nargs, ARG_LIST args[])
 
     nstring        = 0;
     for (i=0; i<nargs; i++) {
-	if (args[i].is_char) {
-	    fprintf(fout,"  char *_%stmp%d=0;\n", stringArgLenName, nstring);
+	if (args[i].is_char && args[i].has_star) {
+	    fprintf(fout,"  char *_%stmp%d = PETSC_NULLPTR;\n", stringArgLenName, nstring);
 	    nstring++;
 	}
     }
     return 0;
 }
 
-int stringFtoC(FILE *fout, int nargs, ARG_LIST args[])
+int stringFtoC(FILE *fout, int nargs, ARG_LIST args[], TYPE_LIST types[])
 {
     int i, nstring;
 
+    for (i=0; i<nargs; i++) {
+        if (strcmp(types[args[i].type].type,"PetscInt") == 0 && args[i].has_star == 1) {
+            fprintf(fout,"CHKFORTRANNULLINTEGER(%s);\n",args[i].name);
+        } else if (strcmp(types[args[i].type].type,"PetscScalar") == 0 && args[i].has_star == 1) {
+            fprintf(fout,"CHKFORTRANNULLSCALAR(%s);\n",args[i].name);
+        } else if (strcmp(types[args[i].type].type,"PetscReal") == 0 && args[i].has_star == 1) {
+            fprintf(fout,"CHKFORTRANNULLREAL(%s);\n",args[i].name);
+        } else if (strncmp(types[args[i].type].type,"MPI_",4) && SYConfigDBLookup("nativeptr", types[args[i].type].type, NULL, nativePtrList) == 1 && args[i].has_star) {
+            if (args[i].has_star == 2) {
+               fprintf(fout,"PetscBool %s_null = !*(void**) %s ? PETSC_TRUE : PETSC_FALSE;\n",args[i].name,args[i].name);
+            }
+            fprintf(fout,"CHKFORTRANNULLOBJECT(%s);\n",args[i].name);
+        }
+    }
+
     /* For now, only implement the most common string style */
     if (stringStyle != STRING_LENEND) return 0;
 
     nstring = 0;     /* reset the string counter */
     for (i=0; i<nargs; i++) {
-	if (args[i].is_char) {
+	if (args[i].is_char && args[i].has_star == 1) {
 	    /* For each character string, copy to a C string.
 	       Note that we could define the end-of-string as the
 	       point where the string is blank padded (this is what
 	       MPICH does).  But in the general case, those trailing
 	       blanks may be meaningful, so we copy the entire string */
-	    fprintf(fout,"/* insert Fortran-to-C conversion for %s */\n",
-		    args[i].name);
+               fprintf(fout,"/* insert Fortran-to-C conversion for %s */\n",
+		    args[i].name);/*
 	    fprintf(fout,"\
   _%stmp%d = (char*)malloc(%s%d+1);\n\
   if (_%stmp%d) {\n\
@@ -2821,14 +2920,15 @@ int stringFtoC(FILE *fout, int nargs, ARG_LIST args[])
 		    stringArgLenName, nstring, stringArgLenName, nstring,
 		    stringArgLenName, nstring,
 		    stringArgLenName, nstring, stringArgLenName, nstring, args[i].name,
-		    stringArgLenName, nstring);
+		    stringArgLenName, nstring);*/
+                fprintf(fout,"  FIXCHAR(%s,%s%d,_%stmp%d);\n",args[i].name,stringArgLenName,nstring,stringArgLenName,nstring);
 	    nstring++;
 	}
     }
     return 0;
 }
 
-int stringCtoF(FILE *fout, int nargs, ARG_LIST args[])
+int stringCtoF(FILE *fout, int nargs, ARG_LIST args[], TYPE_LIST types[])
 {
     int i, nstring;
 
@@ -2838,29 +2938,38 @@ int stringCtoF(FILE *fout, int nargs, ARG_LIST args[])
     nstring = 0;
     for (i=0; i<nargs; i++) {
 	if (args[i].is_char) {
-	    if (!args[i].is_const) {
+            if (args[i].has_star == 2) {
+                /* if (!args[i].is_const) { does not apply because it is const char ** */
 		/* This is more complex than the Fortran-to-C copy,
 		   since a null must be replaced with blank pads */
 		fprintf(fout,"/* insert C-to-Fortran conversion for %s */\n",
 			args[i].name);
-		fprintf(fout,"\
+                /*	fprintf(fout,"          \
   if (_%stmp%d) {\n\
     int _i;\n\
     for(_i=0; _i<%s%d && _%stmp%d[_i]; _i++) %s[_i]=_%stmp%d[_i];\n\
     while (_i<%s%d) %s[_i++] = ' ';\n\
-    free(_%stmp%d);\n\
   }\n",
 			stringArgLenName, nstring,
 			stringArgLenName, nstring, stringArgLenName, nstring, args[i].name, stringArgLenName, nstring,
 			stringArgLenName, nstring, args[i].name,
-			stringArgLenName, nstring);
-	    }
-	    else {
-		fprintf(fout, "  if (_%stmp%d) { free(_%stmp%d); }\n",
-			stringArgLenName, nstring, stringArgLenName, nstring);
+                  stringArgLenName, nstring);*/
+                fprintf(fout,"*ierr = PetscStrncpy(%s, _%stmp%d, %s%d);\n\
+                        if (*ierr) return;\n\
+                        FIXRETURNCHAR(PETSC_TRUE, %s, %s%d);\n",args[i].name,stringArgLenName,nstring,stringArgLenName,nstring,
+                        args[i].name,stringArgLenName,nstring);
+                /* } */
+            }
+	    if (args[i].has_star == 1) {
+                /*	fprintf(fout, "  if (_%stmp%d) { free(_%stmp%d); }\n",
+                  stringArgLenName, nstring, stringArgLenName, nstring); */
+                fprintf(fout,"  FREECHAR(%s,_%stmp%d);\n",args[i].name,stringArgLenName,nstring);
 	    }
 	    nstring++;
-	}
+	} else if (strncmp(types[args[i].type].type,"MPI_",4) && SYConfigDBLookup("nativeptr", types[args[i].type].type, NULL, nativePtrList) == 1 && args[i].has_star == 2) {
+            fprintf(fout,"// if C routine nullifed the object, we must set to to -2 to indicate null set in Fortran\n");
+            fprintf(fout,"if (! %s_null && !*(void**) %s) * (void **) %s = (void *)-2;\n",args[i].name,args[i].name,args[i].name);
+        }
     }
     return 0;
 }
